from datetime import timedelta

DURATION_HUMAN_SPEC = (
    (1.e-6, 1e9, 1e3, 'ns'),
    (1.e-3, 1e6, 1e3, 'us'),
    (1., 1e3, 1e3, 'ms'),
    (60., 1e0, 60., 's'),
)


def duration_human(value):
    """Return a beautiful representation of the duration.
    It dynamically calculates the best unit to use.

    Returns:
        str: the duration representation.

    """
    for top, mult, size, unit in DURATION_HUMAN_SPEC:
        if value < top:
            result = round(value * mult, ndigits=2)
            if result < size:
                return '{}{}'.format(result, unit)

    txt = str(timedelta(seconds=float('{:.1f}'.format(value))))
    pos = txt.find('.')
    if pos == -1:
        return txt
    return txt[:pos + 2]


THROUGHPUT_HUMAN_SPEC = (
    (1. / 60 / 24, 60 * 60 * 24, 24, '/d'),
    (1. / 60, 60 * 60, 60, '/h'),
    (1., 60, 60, '/m'),
    (float('inf'), 1, float('inf'), '/s'),
)


def throughput_human(value):
    """Return a beautiful representation of the current throughput.
    It dynamically calculates the best unit to use.

    Returns:
        str: the duration representation.

    """
    for top, mult, size, unit in THROUGHPUT_HUMAN_SPEC:
        if value < top:
            result = round(value * mult, ndigits=2)
            if result < size:
                return '{}{}'.format(result, unit)
    return '?'



# coding=utf-8
from __future__ import absolute_import, division, unicode_literals

import random
from datetime import datetime
from decimal import Decimal
from itertools import chain, repeat, tee

try:
    from unittest import mock
except ImportError:
    import mock

import pytest

from about_time import about_time, duration_human, throughput_human
from about_time.core import Handle, HandleStats


@pytest.fixture
def rand_offset():
    return random.random() * 1000


@pytest.fixture
def mock_timer():
    with mock.patch('about_time.core.timer') as mt:
        yield mt


def test_duration_context_manager_mode(rand_offset, mock_timer):
    start, end = 1.4 + rand_offset, 2.65 + rand_offset
    mock_timer.side_effect = start, end

    with about_time() as at:
        pass

    assert at.duration == pytest.approx(end - start)


def test_duration_callable_mode(rand_offset, mock_timer):
    start, end = 1.4 + rand_offset, 2.65 + rand_offset
    mock_timer.side_effect = start, end

    at = about_time(lambda: 1)

    assert at.duration == pytest.approx(end - start)


def test_duration_counter_throughput_mode(rand_offset, mock_timer):
    start, end = 1.4 + rand_offset, 2.65 + rand_offset
    mock_timer.side_effect = start, end

    at = about_time(range(2))
    for _ in at:
        pass

    assert at.duration == pytest.approx(end - start)


@pytest.mark.parametrize('call, args, kwargs, expected', [
    (lambda: 123, (), {}, 123),
    (str, (), {}, ''),
    (list, (), {}, []),
    (lambda x: x + 1, (123,), {}, 124),
    (str, ('cool',), {}, 'cool'),
    (list, ((1, 2, 3),), {}, [1, 2, 3]),
    (lambda x: x + 1, (), {'x': 123}, 124),
])
def test_callable_mode_result(call, args, kwargs, expected):
    at = about_time(call, *args, **kwargs)
    assert at.result == expected


@pytest.mark.parametrize('it', [
    [],
    [1, 2, 3],
    range(0),
    range(12),
    'string',
    (x ** 2 for x in range(8)),
])
def test_counter_throughput_mode(it, rand_offset, mock_timer):
    start, end = 1.4 + rand_offset, 2.65 + rand_offset
    mock_timer.side_effect = chain((start,), repeat(end))
    it_see, it_copy = tee(it)

    at = about_time(it_see)
    assert at.count == 0  # count should work even before starting iterating.

    i = 0
    for i, elem in enumerate(at, 1):
        assert elem == next(it_copy)
        assert at.count == i  # count works in real time now!
        assert at.duration > 0  # ensure the timing ending is also updated in real time.

    assert at.throughput == pytest.approx(i / 1.25)


@pytest.mark.parametrize('field', [
    'result',
    'count',
    'throughput',
    'throughput_human',
])
def test_context_manager_mode_dont_have_field(field):
    with about_time() as at:
        pass

    with pytest.raises(AttributeError):
        getattr(at, field)


@pytest.mark.parametrize('field', [
    'count',
    'throughput',
    'throughput_human',
])
def test_callable_mode_dont_have_field(field):
    at = about_time(lambda: 1)

    with pytest.raises(AttributeError):
        getattr(at, field)


@pytest.mark.parametrize('field', [
    'result',
])
def test_counter_throughput_mode_dont_have_field(field):
    at = about_time(range(2))

    with pytest.raises(AttributeError):
        getattr(at, field)


@pytest.mark.parametrize('value', [
    123,
    .1,
    object(),
    datetime.now(),
    Decimal(),
])
def test_wrong_params_must_complain(value):
    with pytest.raises(UserWarning):
        about_time(value)


def test_handle_duration_human():
    h = Handle([1, 2])
    with mock.patch('about_time.human.duration_human') as mdh:
        assert h.duration_human  # the mock is returned.
    mdh.assert_called_once_with(1)


def test_handle_throughput_human():
    def it_closure():
        pass

    it_closure.count = 1
    h = HandleStats([1, 2], it_closure)

    with mock.patch('about_time.human.throughput_human') as mth:
        assert h.throughput_human  # the mock is returned.
    mth.assert_called_once_with(1)


@pytest.mark.parametrize('duration, expected', [
    (.00000000123, '1.23ns'),
    (.00000000185, '1.85ns'),
    (.000000001855, '1.85ns'),
    (.0000000018551, '1.86ns'),
    (.000001, '1.0us'),
    (.000000999996, '1.0us'),
    (.00001, '10.0us'),
    (.0000156, '15.6us'),
    (.01, '10.0ms'),
    (.0141233333333, '14.12ms'),
    (.0199999, '20.0ms'),
    (.1099999, '110.0ms'),
    (.1599999, '160.0ms'),
    (.8015, '801.5ms'),
    (3.434999, '3.43s'),
    (3.435999, '3.44s'),
    (59.99, '59.99s'),
    (59.999, '0:01:00'),
    (60.0, '0:01:00'),
    (68.5, '0:01:08.5'),
    (68.09, '0:01:08.1'),
    (60.99, '0:01:01'),
    (125.825, '0:02:05.8'),
    (4488.395, '1:14:48.4'),
])
def test_duration_human(duration, expected):
    assert duration_human(duration) == expected


@pytest.mark.parametrize('duration, count, expected', [
    (float('nan'), 1, '?'),
    (1., 1, '1.0/s'),
    (1., 10, '10.0/s'),
    (1., 2500, '2500.0/s'),
    (1., 1825000, '1825000.0/s'),
    (5.47945205e-07, 1, '1825000.0/s'),
    (2., 1, '30.0/m'),
    (2., 10, '5.0/s'),
    (2., 11, '5.5/s'),
    (1.981981981981982, 11, '5.55/s'),
    (100., 10, '6.0/m'),
    (100., 3, '1.8/m'),
    (110., 8, '4.36/m'),
    (1600., 3, '6.75/h'),
    (67587655435., 5432737542, '4.82/m'),
    (67587655435., 543273754, '28.94/h'),
    (67587655435., 543273754271, '8.04/s'),
    (.99, 1, '1.01/s'),
    (.999, 1, '1.0/s'),
    (1.00001, 1, '1.0/s'),
    (1.0001, 1, '59.99/m'),
    (1165263., 123, '9.12/d'),
    (3600., 1, '1.0/h'),
    (3601., 1, '23.99/d'),
    (80000., 2, '2.16/d'),
])
def test_throughput_human(duration, count, expected):
    assert throughput_human(count / duration) == expected



# coding=utf-8
from __future__ import absolute_import, division, unicode_literals

import sys
import time
from contextlib import contextmanager

from . import human

if sys.version_info >= (3, 3):
    timer = time.perf_counter
else:  # pragma: no cover
    timer = time.time


def about_time(func_or_it=None, *args, **kwargs):
    """Measure timing and throughput of code blocks, with beautiful
    human friendly representations.

    There're three modes of operation: context manager, callable and
    throughput.

    1. Use it like a context manager:

    >>> with about_time() as at:
    ....    # code block.

    2. Use it with a callable:

    >>> at = about_time(func, arg1, kwarg2=arg2)  # send arguments at will.

    3. Use it with an iterable or generator:

    >>> at = about_time(items)
    >>> for item in at:
    ....    # use item
    """

    timings = [0.0, 0.0]

    # use as a context manager.
    if func_or_it is None:
        return _context_timing(timings, Handle(timings))

    # use as a callable.
    if callable(func_or_it):
        with _context_timing(timings):
            result = func_or_it(*args, **kwargs)
        return HandleResult(timings, result)

    try:
        it = iter(func_or_it)
    except TypeError:
        raise UserWarning('param should be callable or iterable.')

    # use as a counter/throughput iterator.
    def it_closure():
        with _context_timing(timings):
            for it_closure.count, elem in enumerate(it, 1):  # iterators are iterable.
                yield elem

    it_closure.count = 0  # the count will only be updated after starting iterating.
    return HandleStats(timings, it_closure)


@contextmanager
def _context_timing(timings, handle=None):
    timings[0] = timer()
    yield handle
    timings[1] = timer()


class Handle(object):
    def __init__(self, timings):
        self.__timings = timings

    @property
    def duration(self):
        """Return the actual duration in seconds.
        This is dynamically updated in real time.

        Returns:
            float: the number of seconds.

        """
        return (self.__timings[1] or timer()) - self.__timings[0]

    @property
    def duration_human(self):
        """Return a beautiful representation of the duration.
        It dynamically calculates the best unit to use.

        Returns:
            str: the duration representation.

        """
        return human.duration_human(self.duration)


class HandleResult(Handle):
    def __init__(self, timings, result):
        super(HandleResult, self).__init__(timings)
        self.__result = result

    @property
    def result(self):
        """Return the result of the callable.

        Returns:
            the result of the callable.

        """
        return self.__result


class HandleStats(Handle):
    def __init__(self, timings, it_closure):
        super(HandleStats, self).__init__(timings)
        self.__it = it_closure

    def __iter__(self):
        return self.__it()

    @property
    def count(self):
        """Return the current iteration count.
        This is dynamically updated in real time.

        Returns:
            int: the current iteration count.

        """
        return self.__it.count

    @property
    def throughput(self):
        """Return the current throughput in items per second.
        This is dynamically updated in real time.

        Returns:
            float: the number of items per second.

        """
        try:
            return self.count / self.duration
        except ZeroDivisionError:  # pragma: no cover
            return float('nan')

    @property
    def throughput_human(self):
        """Return a beautiful representation of the current throughput.
        It dynamically calculates the best unit to use.

        Returns:
            str: the duration representation.

        """
        return human.throughput_human(self.throughput)
